/* continuation.h                                              -*- C++ -*-
 * Morgan Deters <mdeters@morgandeters.com>
 * September 2003
 *
 * This implements continuations for gcc on x86 platforms with
 * glibc 2.3.2 (mileage may vary on other libc versions).
 *
 * THIS CODE USES GLOBAL VARIABLES AND IS EXPLICITLY NOT THREAD-SAFE.
 * It was written to support a user-space threading library
 * implementation running in a single kernel thread.  In that context,
 * its thread-safety is not an issue since all continuation activity
 * will be atomic with respect to that kernel thread, and none of the
 * user-space threads it manages will access the continuation library
 * (or its global state) directly.
 *
 * This file is part of the RMA Templates project.  For additional
 * information and build instructions, please see the README file in
 * the top-level source directory.
 *
 * Copyright (c) 2003 MORGAN DETERS
 * All rights reserved.
 *
 * This software is open-source software and may be redistributed and
 * modified under the terms of an amended BSD License.  For the terms
 * of the license, see the file COPYING in the top-level directory.
 * If this file is missing for some reason, please obtain the latest
 * release of these sources from:
 *   http://www.cs.wustl.edu/~mdeters/projects/middleware/#rma_templates
 */

#ifndef CONTINUATION_H
#define CONTINUATION_H

#if __cplusplus
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <csetjmp>
#include <new>
#else
#error C++ required for continuation library (for now)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#endif /* __cplusplus */

#include <sys/resource.h>
#include <unistd.h>

#define DEBUG(x) printf x
//#define DEBUG(x)

namespace continuation {
  class continuation_t;
}

// one MiB
#define CONTINUATION_STACK_BYTES (1024*512)

class ::continuation::continuation_t {
  static continuation_t *cc_;      /* current continuation */

  sigjmp_buf self_;                /* for siglongjmp() */
  int arg_;                        /* continuation user arg */
  int n_;                          /* the # of bytes of stack */
  char *stack_;                    /* the continuation's stack */
  union {
    void (*entry_ptr_)(continuation_t*);
    void (*entry_ref_)(continuation_t&);
    void (*entry_void_)();
  } e_;

  static void enter_cc(continuation_t*) throw();
  static void enter(continuation_t*) throw();

  inline continuation_t() throw();

public:
  inline continuation_t(void (*)(continuation_t*)) throw(std::bad_alloc);
  inline continuation_t(void (*)(continuation_t&)) throw(std::bad_alloc);
  inline continuation_t(void (*)()) throw(std::bad_alloc);
  inline ~continuation_t() throw();

  int operator()(int) throw();
  inline operator bool() const throw();
  void jump(int) throw() __attribute__((__noreturn__));
  inline int call(int) throw();
  bool save() throw();
  inline int arg() const throw();

  inline static continuation_t* get() throw();
};

#define set_stack_unlimited(rlim) ({ \
  if(getrlimit(RLIMIT_STACK, &(rlim)) == -1) { \
    perror("getrlimit"); \
    abort(); \
  } \
  (rlim).rlim_cur = (rlim).rlim_max; \
  if(setrlimit(RLIMIT_STACK, &(rlim)) == -1) { \
    perror("setrlimit"); \
    abort(); \
  } \
})

#define set_stack_limited_to(rlim, stack_n) ({ \
  if(getrlimit(RLIMIT_STACK, &(rlim)) == -1) { \
    perror("getrlimit"); \
    abort(); \
  } \
  if((stack_n) == 0) \
    (rlim).rlim_cur = (rlim).rlim_max; \
  else (rlim).rlim_cur = (stack_n); \
  DEBUG(("limit to %u[%u]\n", (stack_n), (rlim).rlim_cur)); \
  if(setrlimit(RLIMIT_STACK, &(rlim)) == -1) { \
    perror("setrlimit"); \
    abort(); \
  } \
})

inline ::continuation::continuation_t::continuation_t() throw() {
}

#define SETUP_STACK \
  stack_ = new char[n_]; \
  save(); \
  int sp = (int)(stack_ + n_); \
  self_[0].__jmpbuf[JB_SP] = sp - 4; \
  *(int*)sp = (int)get(); \
  self_[0].__jmpbuf[JB_PC] = (int)

inline ::continuation::continuation_t::continuation_t(void (*e)(continuation_t*)) throw(std::bad_alloc)
  : n_(CONTINUATION_STACK_BYTES) {
  e_.entry_ptr_ = e;
  SETUP_STACK enter_cc;
}

inline ::continuation::continuation_t::continuation_t(void (*e)(continuation_t&)) throw(std::bad_alloc)
  : n_(CONTINUATION_STACK_BYTES) {
  e_.entry_ref_ = e;
  SETUP_STACK enter_cc;
}

inline ::continuation::continuation_t::continuation_t(void (*e)()) throw(std::bad_alloc)
  : n_(CONTINUATION_STACK_BYTES) {
  e_.entry_void_ = e;
  SETUP_STACK enter;
}

#undef SETUP_STACK

inline ::continuation::continuation_t::~continuation_t() throw() {
  delete[] stack_;
}

inline ::continuation::continuation_t::operator bool() const throw() {
  return this == cc_;
}

inline int ::continuation::continuation_t::arg() const throw() {
  return arg_;
}

inline ::continuation::continuation_t* ::continuation::continuation_t::get() throw() {
  if(cc_ == 0)
    cc_ = new continuation_t;
  return cc_;
}

inline int ::continuation::continuation_t::call(int i) throw() {
  return (*this)(i);
}

#undef set_stack_unlimited
#undef set_stack_limited_to

#endif /* CONTINUATION_H */
