/* continuation_scheduler.h                                    -*- C++ -*-
 * Non-threading scheduler for RMA Template schedules
 * x86 glibc 2.3.2 version (THIS SCHEDULER IS NOT PORTABLE OR THREAD-SAFE)
 *
 * This file is part of the RMA Templates project.  For additional
 * information and build instructions, please see the README file in
 * the top-level source directory.
 *
 * Copyright (c) 2003, MORGAN DETERS, CHRISTOPHER GILL, and RON CYTRON
 * All rights reserved.
 *
 * This software is open-source software and may be redistributed and
 * modified under the terms of an amended BSD License.  For the terms
 * of the license, see the file COPYING in the top-level directory.
 * If this file is missing for some reason, please obtain the latest
 * release of these sources from:
 *   http://www.cs.wustl.edu/~mdeters/projects/middleware/#rma_templates
 */

#include <cstdlib>
#include <csignal>

#include <unistd.h>
#include <sys/time.h>

#include <loki/NullType.h>
#include <loki/Typelist.h>

#include <loki/static_check.h>

#include "continuation.h"


/*
#include <iostream>
#include <iomanip>
#define DEBUG(x) ({ struct timeval __tv; \
                    if(gettimeofday(&__tv, 0) == -1) \
                      perror("gettimeofday"); \
                    std::cout << __tv.tv_sec << '.' \
                              << std::setw(3) \
                              << (__tv.tv_usec / 1000) << ' ' \
                              << x << std::endl; 1; })
*/
#define DEBUG(x)

namespace RMA {
  template <class Task>
  struct Continuation_Task {
    static continuation::continuation_t *ctn_, *next_;
    static unsigned t_;
    static bool finished_;
    static void start(continuation::continuation_t* next) {
      DEBUG("STARTING TASK");
      setup_timer(t_);
      next_ = next;
      Task::entry();
      for(;;) {
        finished_ = true;
        DEBUG("FINISHED TASK - Zzzzzzzzz...");
        sleep(300);
      }
    }
    static void timeout(int sig) {
      DEBUG("TIMEOUT -- continuing into scheduler");
      (void) (*next_)(0);
      DEBUG("TIMEOUT RETURN -- continuing into task (timer == " << t_ << ')');
      setup_timer(t_);
    }

    static void setup_timer(unsigned t) {
      unsigned sec = t / RMS_HZ;
      unsigned usec = (t % RMS_HZ) * (1000000.0 / RMS_HZ);

      struct itimerval the_timer;
      struct sigaction act;

      the_timer.it_interval.tv_sec  = the_timer.it_value.tv_sec  = sec;
      the_timer.it_interval.tv_usec = the_timer.it_value.tv_usec = usec;

      act.sa_handler = timeout;
      act.sa_flags = SA_ONESHOT /* | SA_NODEFER */ ;

      if(sigaction(SIGALRM, &act, 0) == -1)
        perror("sigaction");

      if(setitimer(ITIMER_REAL, &the_timer, 0) == -1)
        perror("setitimer");

      DEBUG("set timer to " << sec << "s " << usec << "us");
    }

    static void release_suspend(unsigned t) {
      ctn_ = new continuation::continuation_t(Continuation_Task<Task>::start);
      t_ = t;
      finished_ = false;
      next_ = 0;
      DEBUG("RELEASE -- continuing into task");
      (void) (*ctn_)((int) &ctn_);
      DEBUG("RELEASE/return -- back in scheduler");
    }
    static void release_finish(unsigned t) {
      ctn_ = new continuation::continuation_t(Continuation_Task<Task>::start);
      t_ = t;
      finished_ = false;
      next_ = 0;
      DEBUG("RELEASE -- continuing into task");
      (void) (*ctn_)((int) &ctn_);
      DEBUG("RELEASE/return -- back in scheduler, deleting ctn, finished_ == " << finished_);
      if(!finished_)
        std::cerr << "TASK OVERRAN BUDGET" << std::endl;
      delete ctn_;
      ctn_ = 0;
    }
    static void resume_suspend(unsigned t) {
      t_ = t;
      DEBUG("RESUME -- continuing into task");
      (void) (*ctn_)((int) &ctn_);
      DEBUG("RESUME/return -- back in scheduler");
    }
    static void resume_finish(unsigned t) {
      t_ = t;
      DEBUG("RESUME -- continuing into task");
      (void) (*ctn_)((int) &ctn_);
      DEBUG("RESUME/return -- back in scheduler, deleting ctn, finished_ == " << finished_);
      if(!finished_)
        std::cerr << "TASK OVERRAN BUDGET" << std::endl;
      delete ctn_;
      ctn_ = 0;
    }
  };

  template <class Task>
  continuation::continuation_t *Continuation_Task<Task>::ctn_ = 0;

  template <class Task>
  continuation::continuation_t *Continuation_Task<Task>::next_ = 0;

  template <class Task>
  unsigned Continuation_Task<Task>::t_ = 0;

  template <class Task>
  bool Continuation_Task<Task>::finished_ = false;

  template <class TaskSet, unsigned lcm>
  struct Continuation_Scheduler_Expand;

  template <unsigned t1, unsigned t2, class Task, class Tail, unsigned lcm>
  struct Continuation_Scheduler_Expand<Loki::Typelist<Release<t1, Task>,
                                       Loki::Typelist<Suspend<t2, Task>,
                                                      Tail> >, lcm> {
    static void schedule(void) {
      DEBUG(t1 << " RELEASE/SUSPEND " << t2);
      Continuation_Task<Task>::release_suspend(t2 - t1);
      Continuation_Scheduler_Expand<Tail, lcm>::schedule();
    }
  };

  template <unsigned t1, unsigned t2, class Task, class Tail, unsigned lcm>
  struct Continuation_Scheduler_Expand<Loki::Typelist<Release<t1, Task>,
                                       Loki::Typelist<Finish<t2, Task>,
                                                      Tail> >, lcm> {
    static void schedule(void) {
      DEBUG(t1 << " RELEASE/FINISH " << t2);
      Continuation_Task<Task>::release_finish(t2 - t1);
      Continuation_Scheduler_Expand<Tail, lcm>::schedule();
    }
  };

  template <unsigned t1, unsigned t2, class Task, class Tail, unsigned lcm>
  struct Continuation_Scheduler_Expand<Loki::Typelist<Resume<t1, Task>,
                                       Loki::Typelist<Suspend<t2, Task>,
                                                      Tail> >, lcm> {
    static void schedule(void) {
      DEBUG(t1 << " RESUME/SUSPEND " << t2);
      Continuation_Task<Task>::resume_suspend(t2 - t1);
      Continuation_Scheduler_Expand<Tail, lcm>::schedule();
    }
  };

  template <unsigned t1, unsigned t2, class Task, class Tail, unsigned lcm>
  struct Continuation_Scheduler_Expand<Loki::Typelist<Resume<t1, Task>,
                                       Loki::Typelist<Finish<t2, Task>,
                                                      Tail> >, lcm> {
    static void schedule(void) {
      DEBUG(t1 << " RESUME/FINISH " << t2);
      Continuation_Task<Task>::resume_finish(t2 - t1);
      Continuation_Scheduler_Expand<Tail, lcm>::schedule();
    }
  };

  template <unsigned t, unsigned n, class Tail, unsigned lcm>
  struct Continuation_Scheduler_Expand<Loki::Typelist<Idle<t, n>, Tail>, lcm> {
    static void schedule(void) {
      DEBUG(t << " IDLE(" << n << ')');
      // idle();
      Continuation_Scheduler_Expand<Tail, lcm>::schedule();
    }
  };

  template <unsigned lcm>
  struct Continuation_Scheduler_Expand<Loki::NullType, lcm> {
    static void schedule(void) {
    }
  };

  template <class TaskSet, unsigned lcm>
  struct Continuation_Scheduler {
    static void schedule(void) {
      for(;;) {
        DEBUG("MAJOR CYCLE BEGINNING lcm = " << lcm);
        Continuation_Scheduler_Expand<TaskSet, lcm>::schedule();
      }
    }
  };
}
