/* rma.h                                                       -*- C++ -*-
 * Definition of templates to determine taskset feasibility

 * This is a template metaprogram implementing the RMA feasibility
 * test of:
 *   J.P. Lehoczky, L. Sha, and Y. Ding, "The rate monotonic
 *   scheduling algorithm---exact characterization and average-case
 *   behaviour," IEEE Real-Time Systems Symposium (RTSS), Dec 1989.
 * See also:
 *   L. Sha and J.B. Goodenough, "Real-time scheduling theory and
 *   Ada," IEEE Computer, 23(4):53--62, 1990.
 * For additional discussion of this template metaprogram
 * implementation of the test, see:
 *   Morgan Deters, Christopher Gill, and Ron Cytron, "Rate-monotonic
 *   analysis in the C++ type system," RTAS workshop on Model-Driven
 *   Embedded Systems (MDES), May 2003.
 *   http://www.cs.wustl.edu/~mdeters/projects/middleware/#rma_templates
 *
 * This file is part of the RMA Templates project.  For additional
 * information and build instructions, please see the README file in
 * the top-level source directory.
 *
 * Copyright (c) 2003, MORGAN DETERS, CHRISTOPHER GILL, and RON CYTRON
 * All rights reserved.
 *
 * This software is open-source software and may be redistributed and
 * modified under the terms of an amended BSD License.  For the terms
 * of the license, see the file COPYING in the top-level directory.
 * If this file is missing for some reason, please obtain the latest
 * release of these sources from:
 *   http://www.cs.wustl.edu/~mdeters/projects/middleware/#rma_templates
 */

#include <loki/NullType.h>
#include <loki/Typelist.h>

namespace RMA {
  template <class TL, int i, int t, int j = 0>
  struct sum_j {
    typedef typename Loki::TL::TypeAt<TL, j>::Result J;
    enum { Cj = J::cost,
           Tj = J::period,
           my_result = Cj * ((t % Tj > 0 ? 1 : 0) + (t / Tj)),
           Result = sum_j<TL, i, t, j + 1>::Result + my_result };
  };

  template <class TL, int i, int t>
  struct sum_j<TL, i, t, i> {
    enum { Result = 0 };
  };

  template <class TL, int i, int t_ix, int k = 0>
  struct get_t {
    enum { Ti = Loki::TL::TypeAt<TL, i - 1>::Result::period,
           Tk = Loki::TL::TypeAt<TL, k>::Result::period,
           num_l = Ti / Tk,
           Result = (t_ix >= num_l)
           ? get_t<TL, i, t_ix - num_l, k + 1>::Result
           : (t_ix + 1) * Tk };
  };

  template <class TL, int i, int t_ix>
  struct get_t<TL, i, t_ix, i> {
    enum { Result = 0 };
  };

  template <class TL, int i, int t_ix = 0>
  struct task_feasible {
    typedef get_t<TL, i, t_ix> t_type;
    enum { t = t_type::Result,
           Result = (t > 0) &&
           ( sum_j<TL, i, t>::Result <= t
             || task_feasible<TL, i, t_ix + 1>::Result ) };
  };

  template <class TL, int i>
  struct task_feasible<TL, i, i> {
    enum { Result = 0 };
  };

  template <class TL, int m, int i>
  struct check_i;

  template <class Head, class Tail, int m, int i>
  struct check_i<Loki::Typelist<Head, Tail>, m, i> {
    enum { task_result = task_feasible<Loki::Typelist<Head, Tail>, i>::Result,
           Result = check_i<Loki::Typelist<Head, Tail>, m, i + 1>::Result
           && task_result };
  };

  template <class Head, class Tail, int m>
  struct check_i<Loki::Typelist<Head, Tail>, m, m> {
    enum { Result = task_feasible<Loki::Typelist<Head, Tail>, m>::Result };
  };

  template <class TaskSet>
  struct RMA_Feasible {
    enum { m = Loki::TL::Length<TaskSet>::value,
           Result = check_i<TaskSet, m, 1>::Result };
  };
}
