/* schedule.h                                                  -*- C++ -*-
 * Definition of Schedule<> template and supporting code
 *
 * This file is part of the RMA Templates project.  For additional
 * information and build instructions, please see the README file in
 * the top-level source directory.
 *
 * Copyright (c) 2003, MORGAN DETERS, CHRISTOPHER GILL, and RON CYTRON
 * All rights reserved.
 *
 * This software is open-source software and may be redistributed and
 * modified under the terms of an amended BSD License.  For the terms
 * of the license, see the file COPYING in the top-level directory.
 * If this file is missing for some reason, please obtain the latest
 * release of these sources from:
 *   http://www.cs.wustl.edu/~mdeters/projects/middleware/#rma_templates
 */

#include <loki/NullType.h>
#include <loki/Typelist.h>

#include "null_scheduler.h"

/* Frequency of the RMS time clock (should be an integer).
 *
 * For example, RMS_HZ == 1 means that a task with period 2 executes
 * once every two seconds.  A higher RMS_HZ setting allows finer-grain
 * schedules, but a value beyond (or even just close enough to) the
 * frequency of the system clock or scheduler will cause problems.  In
 * Linux, this means on the order of ~10ms (RMS_HZ == 100).
 *
 * Currently this is a global setting.  In the future it will be a
 * per-schedule or per-task parameter, and this #define may specify a
 * default. */
#define RMS_HZ 20

namespace RMA {
  template <class TaskSet>
  struct Major_Cycle;

  template <class TaskSet, class Cycle = Loki::NullType>
  struct TaskSet_to_Cycle_Calc;

  template <class Task, class Cycle>
  struct Insert_into_Cycle_Calc;

  template <unsigned release, class Task>
  struct Pending_Release;

  template <class Cycle>
  struct Build_Major_Cycle;

  template <class ReleaseTable, class Exec, class PendList>
  struct Execute_Major_Cycle;

  template <class Cycle, unsigned lcm>
  struct Build_Release_Table;

  template <class TaskSet>
  struct LCM_Period;

  template <unsigned t, class ReleaseTable>
  struct Find_Next_Task;

  template <bool gt, unsigned t, class First, class Second, class Tail>
  struct Find_Next_Task_helper;

  template <class TaskSet, template <class, unsigned> class Scheduler_T = Null_Scheduler>
  struct Schedule;

  template <unsigned a, unsigned b> struct GCD;
  template <unsigned a, unsigned b> struct LCM;

  template <class Head, class Tail, template <class, unsigned> class Scheduler_T>
  struct Schedule<Loki::Typelist<Head, Tail>, Scheduler_T> {
    typedef Loki::Typelist<Head, Tail> TaskSet;
    enum { feasible = RMA_Feasible<TaskSet>::Result };
    static const double utilization = Schedule<Tail>::utilization +
                                      double(Head::cost) / Head::period;
    static void schedule(void) {
      typedef Major_Cycle<TaskSet> Cycle;
      typedef Scheduler_T<typename Cycle::Result, Cycle::lcm> Scheduler;
      Scheduler::schedule();
    }
  };

  template <template <class, unsigned> class Scheduler_T>
  struct Schedule<Loki::NullType, Scheduler_T> {
    typedef Loki::NullType TaskSet;
    enum { feasible = 1 };
    static const double utilization = 0.0;
    static void schedule(void) {
      // no action necessary
    }
  };

  template <class TaskSet>
  struct Major_Cycle {
    typedef typename TaskSet_to_Cycle_Calc<TaskSet>::Result Cycle_Calc;
    enum { lcm = LCM_Period<TaskSet>::value };
    typedef typename Build_Release_Table<Cycle_Calc, lcm>::Result Releases;
    typedef typename Build_Major_Cycle<Releases>::Result Result;
  };

  template <class THead, class TTail, class Cycle>
  struct TaskSet_to_Cycle_Calc<Loki::Typelist<THead, TTail>, Cycle> {
    typedef Pending_Release<THead::release, THead> Pending;
    typedef typename Insert_into_Cycle_Calc<Pending, Cycle>::Result New_Cycle;
    typedef typename TaskSet_to_Cycle_Calc<TTail, New_Cycle>::Result Result;
  };

  template <class Cycle>
  struct TaskSet_to_Cycle_Calc<Loki::NullType, Cycle> {
    typedef Cycle Result;
  };

  template <class CHead, class CTail, class Task>
  struct Insert_into_Cycle_Calc<Task, Loki::Typelist<CHead, CTail> > {
    // sort them by (release,period)
    typedef typename Loki::Select<
        (CHead::release > Task::release ||
          (CHead::release == Task::release &&
           CHead::Task::period >= Task::Task::period)),
      Loki::Typelist<Task, Loki::Typelist<CHead, CTail> >,
      Loki::Typelist<CHead, typename Insert_into_Cycle_Calc<Task, CTail>::Result> >::Result Result;
  };

  template <class Task>
  struct Insert_into_Cycle_Calc<Task, Loki::NullType> {
    typedef Loki::Typelist<Task, Loki::NullType> Result;
  };

  template <unsigned release_, class Task_>
  struct Pending_Release {
    typedef Task_ Task;
    enum { release = release_ };
  };

  template <class Head, class Tail>
  struct LCM_Period<Loki::Typelist<Head, Tail> > {
    enum { value = LCM<Head::period, LCM_Period<Tail>::value>::value };
  };

  template <class Head>
  struct LCM_Period<Loki::Typelist<Head, Loki::NullType> > {
    enum { value = Head::period };
  };

  template <bool geq, class CHead, class CTail, unsigned lcm>
  struct Build_Release_Table_builder;
  template <class CHead, class CTail, unsigned lcm>
  struct Build_Release_Table_builder<true, CHead, CTail, lcm> {
    typedef Loki::NullType Result;
  };
  template <class CHead, class CTail, unsigned lcm>
  struct Build_Release_Table_builder<false, CHead, CTail, lcm> {
    typedef Pending_Release<CHead::release + CHead::Task::period,
                            typename CHead::Task> Pending;
    typedef typename Insert_into_Cycle_Calc<Pending, CTail>::Result New_Tail;
    typedef typename Build_Release_Table<New_Tail, lcm>::Result Next;

    typedef Loki::Typelist<CHead, Next> Result;
  };

  template <class CHead, class CTail, unsigned lcm>
  struct Build_Release_Table<Loki::Typelist<CHead, CTail>, lcm> {
    typedef typename Build_Release_Table_builder<(CHead::release >= lcm), CHead, CTail, lcm>::Result Result;
  };

  template <unsigned lcm>
  struct Build_Release_Table<Loki::NullType, lcm> {
    typedef Loki::NullType Result;
  };

  template <unsigned t_, class Task_>
  struct Release {
    enum { t = t_ };
    typedef Task_ Task;
  };
  template <unsigned t_, class Task_>
  struct Suspend {
    enum { t = t_ };
    typedef Task_ Task;
  };
  template <unsigned t_, class Task_>
  struct Resume {
    enum { t = t_ };
    typedef Task_ Task;
  };
  template <unsigned t_, class Task_>
  struct Finish {
    enum { t = t_ };
    typedef Task_ Task;
  };
  template <unsigned t_, unsigned n_>
  struct Idle {
    enum { t = t_ };
    typedef struct {
      enum { release = t_,
             period = 0,
             cost = n_
      };
    } Task;
  };

  template <unsigned t_, unsigned rc_, class Task_>
  struct Executing {
    enum { t = t_, rc = rc_ };
    typedef Task_ Task;
  };
  template <unsigned rc_, class Task_>
  struct Pending {
    enum { rc = rc_ };
    typedef Task_ Task;
  };

  template <class RHead, class RTail>
  struct Build_Major_Cycle<Loki::Typelist<RHead, RTail> > {
    typedef Execute_Major_Cycle<RTail,
                                Executing<RHead::release,
                                          RHead::Task::cost,
                                          typename RHead::Task>,
                                Loki::NullType> Exec;
    typedef Loki::Typelist<Release<RHead::release, typename RHead::Task>,
                           typename Exec::Result> Result;
  };

  template <class ReleaseTable, class Exec, class PendList = Loki::NullType>
  struct Execute_Major_Cycle_Suspend_Release {
    enum { finish_time = Exec::t + Exec::rc };
    typedef Find_Next_Task<finish_time, ReleaseTable> FNT;
    typedef typename FNT::Task Next_Task;
    typedef typename FNT::ReleaseTable New_ReleaseTable;
    typedef Execute_Major_Cycle<New_ReleaseTable,
                                Executing<Next_Task::release, Next_Task::Task::cost, typename Next_Task::Task>,
                                Loki::Typelist<Pending<finish_time - Next_Task::release, typename Exec::Task>, PendList> > Next;
    typedef Loki::Typelist<Suspend<Next_Task::release, typename Exec::Task>,
            Loki::Typelist<Release<Next_Task::release, typename Next_Task::Task>,
                           typename Next::Result> > Result;
  };

  template <unsigned t, class First, class Second, class Tail>
  struct Find_Next_Task_helper<true, t, First, Second, Tail> {
    typedef First Task;
    typedef Loki::Typelist<Second, Tail> ReleaseTable;
  };

  template <unsigned t, class First, class Second, class Tail>
  struct Find_Next_Task_helper<false, t, First, Second, Tail> {
    typedef Find_Next_Task<t, Loki::Typelist<Second, Tail> > contn;
    enum { we_win = First::Task::period <= contn::Task::Task::period };
    typedef typename Loki::Select<we_win, First, typename contn::Task>::Result Task;
    typedef typename Loki::Select<we_win,
                                  Loki::Typelist<Second, Tail>,
                                  Loki::Typelist<First, typename contn::ReleaseTable> >::Result ReleaseTable;
  };

  template <unsigned t, class First, class Second, class Tail>
  struct Find_Next_Task<t, Loki::Typelist<First, Loki::Typelist<Second, Tail> > > {
    enum { exists = 1 };
    typedef Find_Next_Task_helper<(Second::release > t), t, First, Second, Tail> FNTh;
    typedef typename FNTh::Task Task;
    typedef typename FNTh::ReleaseTable ReleaseTable;
  };

  template <unsigned t, class First>
  struct Find_Next_Task<t, Loki::Typelist<First, Loki::NullType> > {
    enum { exists = 1 };
    typedef First Task;
    typedef Loki::NullType ReleaseTable;
  };

  template <unsigned t>
  struct Find_Next_Task<t, Loki::NullType> {
    enum { exists = 0 };
    typedef void Task;
    typedef void ReleaseTable;
  };

  template <class ReleaseTable, class Exec, class PendList = Loki::NullType>
  struct Execute_Major_Cycle_Finish_Release {
    enum { finish_time = Exec::t + Exec::rc };
    typedef Find_Next_Task<finish_time, ReleaseTable> FNT;
    typedef typename FNT::Task Next_Task;
    typedef typename FNT::ReleaseTable New_ReleaseTable;
    enum { next_start_time = (finish_time > Next_Task::release
                              ? finish_time
                              : Next_Task::release),
           idle_time = next_start_time - finish_time };
    typedef Execute_Major_Cycle<New_ReleaseTable,
                                Executing<next_start_time, Next_Task::Task::cost,
                                          typename Next_Task::Task>,
                                PendList> Next;
    typedef Loki::Typelist<Release<next_start_time, typename Next_Task::Task>,
                           typename Next::Result> Tail;
    typedef Loki::Typelist<Finish<finish_time, typename Exec::Task>,
                           typename Loki::Select<(idle_time > 0),
                                                 Loki::Typelist<Idle<finish_time, idle_time>, Tail>,
                                                 Tail>::Result> Result;
  };

  template <class ReleaseTable, class Exec, class PHead, class PTail>
  struct Execute_Major_Cycle_Finish_Resume {
    enum { finish_time = Exec::t + Exec::rc };
    typedef Execute_Major_Cycle<ReleaseTable,
                                Executing<finish_time, PHead::rc, typename PHead::Task>,
                                PTail> Next;
    typedef Loki::Typelist<Finish<finish_time, typename Exec::Task>,
            Loki::Typelist<Resume<finish_time, typename PHead::Task>,
                           typename Next::Result> > Result;
  };

  template <class RHead, class RTail, class Exec>
  struct Execute_Major_Cycle<Loki::Typelist<RHead, RTail>,
                             Exec,
                             Loki::NullType> {
    typedef Loki::Typelist<RHead, RTail> ReleaseTable;

    enum { finish_time = Exec::t + Exec::rc };

    typedef Find_Next_Task<finish_time, ReleaseTable> Candidate_Task;

    typedef typename Loki::Select<
      (Candidate_Task::exists
       && Candidate_Task::Task::Task::period < Exec::Task::period
       && Candidate_Task::Task::release < finish_time),
      Execute_Major_Cycle_Suspend_Release<ReleaseTable, Exec>,
      Execute_Major_Cycle_Finish_Release<ReleaseTable, Exec> >::Result selection;

    typedef typename selection::Result Result;
  };

  template <class RHead, class RTail, class Exec, class PHead, class PTail>
  struct Execute_Major_Cycle<Loki::Typelist<RHead, RTail>,
                             Exec,
                             Loki::Typelist<PHead, PTail> > {
    typedef Loki::Typelist<RHead, RTail> ReleaseTable;
    typedef Loki::Typelist<PHead, PTail> PendList;

    enum { finish_time = Exec::t + Exec::rc };

    typedef Find_Next_Task<finish_time, ReleaseTable> Candidate_Task;

    typedef typename Loki::Select<
      (Candidate_Task::exists
       && Candidate_Task::Task::Task::period < Exec::Task::period
       && Candidate_Task::Task::release < finish_time),
      Execute_Major_Cycle_Suspend_Release<ReleaseTable, Exec, PendList>,
      typename Loki::Select<
        (Candidate_Task::Task::Task::period < PHead::Task::period && Candidate_Task::Task::release <= finish_time),
        Execute_Major_Cycle_Finish_Release<ReleaseTable, Exec, PendList>,
        Execute_Major_Cycle_Finish_Resume<ReleaseTable, Exec, PHead, PTail>
      >::Result>::Result selection;

    typedef typename selection::Result Result;
  };

  template <class Exec, class PHead, class PTail>
  struct Execute_Major_Cycle<Loki::NullType,
                             Exec,
                             Loki::Typelist<PHead, PTail> > {
    enum { finish_time = Exec::t + Exec::rc };
    typedef Executing<finish_time, PHead::rc, typename PHead::Task> New_Exec;
    typedef Execute_Major_Cycle<Loki::NullType, New_Exec, PTail> Next;
    typedef Loki::Typelist<Finish<finish_time, typename Exec::Task>,
            Loki::Typelist<Resume<finish_time, typename PHead::Task>,
                           typename Next::Result> > Result;
  };

  template <class Exec>
  struct Execute_Major_Cycle<Loki::NullType, Exec, Loki::NullType> {
    typedef Loki::Typelist<Finish<Exec::t + Exec::rc, typename Exec::Task>,
                           Loki::NullType> Result;
  };

  template <unsigned a, unsigned b>
  struct GCD {
    enum { value = GCD<b, a % b>::value };
  };

  template <unsigned a>
  struct GCD<a, 0> {
    enum { value = a };
  };

  template <unsigned a, unsigned b>
  struct LCM {
    enum { value = a * b / GCD<a, b>::value };
  };
}
